/*
 *  TestPC.java
 *  Joris van Rantwijk
 */

package retro;
import java.io.*;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

public class TestPC
  implements Cpu.InterruptHook, Cpu.TraceHook, ExternalInputHandler
{

    Scheduler sched;
    Cpu cpu;
    Memory mem;
    IOPorts io;
    IOMisc iomisc;
    Cga cga;
    I8259 intctl;
    I8254 timer;
    KeyboardController kbdctl;
    ExternalInputHandler keyInputHandler;

    DiskImage diskimage;
    TerminalFrame frame;

    private static final void msg(String s)
    {
        System.err.println(s);
    }

    // Intercepted interrupts
    public int interruptHook(int intno, int[] reg, int[] sreg, int[] flags)
    {
        if (intno != 0xe6) {
            System.err.println("int" + Misc.byteToHex(intno) + ": ax=" + Misc.wordToHex(reg[Cpu.regAX]) + " cx=" + Misc.wordToHex(reg[Cpu.regCX]) + " dx=" + Misc.wordToHex(reg[Cpu.regDX]));
            return intno;
        }

        int ax = reg[Cpu.regAX];
        if ((ax >> 8) == 0xfe) {
            int[] fakeflags = new int[1];
            reg[Cpu.regAX] = mem.loadWord(
              (sreg[Cpu.sregSS] << 4) + reg[Cpu.regSP]);
            fakeflags[0] = mem.loadWord(
              (sreg[Cpu.sregSS] << 4) + ((reg[Cpu.regSP] + 6) & 0xffff));
            switch (ax & 0xff) {
              case 0x10: int_video(reg, sreg, fakeflags); break;
              case 0x13: int_disk(reg, sreg, fakeflags); break;
              default: System.err.println("Intercepted unknown interrupt 0x" + Misc.byteToHex(intno)); break;
            }
            mem.storeWord(
              (sreg[Cpu.sregSS] << 4) + reg[Cpu.regSP],
              reg[Cpu.regAX]);
            mem.storeWord(
              (sreg[Cpu.sregSS] << 4) + ((reg[Cpu.regSP] + 6) & 0xffff),
              fakeflags[0]);
            reg[Cpu.regAX] = ax;
        } else {
            System.err.println("Unknown function int 0xe6 ax=" + Misc.wordToHex(ax));
        }
        return -1;
    }

    
    // Write instruction trace
    private int tracecount = -1;
    private OutputStream dumpf;
    public void traceHook()
    {
        if (tracecount < 0) return;
        try {
            if (tracecount == 1) {
                dumpf = new FileOutputStream("dump.start");
                byte[] state = cpu.getStateData();
                dumpf.write(state);
                dumpf.write(mem.mem);
                dumpf.close();
                dumpf = new BufferedOutputStream(
                  new FileOutputStream("dump.trace"));
            }
            if (tracecount > 1) {
                byte[] state = cpu.getStateData();
                dumpf.write(state);
            }
            tracecount++;
        } catch (IOException e) {
            e.printStackTrace();
            System.exit(1);
        }
    }


	// Handle interrupt 0x10
	void int_video(int[] reg, int[] sreg, int[] flags)
	{
		int ax = reg[Cpu.regAX];
		switch ((ax >> 8) & 0xff) {
		  case 0x00: msg("int10: set video mode"); break;
		  case 0x01: msg("int10: set cursor shape"); break;
		  case 0x02: /*msg("int10: set cursor position");*/ break;
		  case 0x03: /*msg("int10: get cursor position");*/ break;
		  case 0x04: msg("int10: get lightpen"); break;
		  case 0x05: msg("int10: set page"); break;
		  case 0x06: msg("int10: scroll up"); break;
		  case 0x07: msg("int10: scroll down"); break;
		  case 0x08: msg("int10: get character"); break;
		  case 0x09: System.out.print((char)(ax & 0xff)); break;
		  case 0x0a: System.out.print((char)(ax & 0xff)); break;
		  case 0x0b: msg("int10: set background"); break;
		  case 0x0e: System.out.print((char)(ax & 0xff)); break;
		  case 0x0f: reg[Cpu.regAX] = 0x5003; reg[Cpu.regBX] &= 0x00ff; break;
		  default: msg("int10: unknown function ax=" + Misc.wordToHex(ax)); break;
		}
	}
	

	int int13_readsectors(int drive, int track, int sect, int head, int n, int addr)
	{
		int ntrack, nsect, nhead = 2;

		long size = diskimage.getSize();
		if (size == 360 * 1024) {
			ntrack = 40;
			nsect = 9;
		} else if (size == 720 * 1024) {
			ntrack = 80;
			nsect = 9;
		} else if (size == 1200 * 1024) {
			ntrack = 80;
			nsect = 15;
		} else if (size == 1440 * 1024) {
			ntrack = 80;
			nsect = 18;
		} else {
			System.err.println("Invalid disk size");
			return 0x180;
		}

		msg("int13: read sectors (drive=" + drive + ", track=" + track + ", sect=" + sect + ", head=" + head + ", n=" + n + ")");

		if (drive != 0)
			return 0x180;

		if (track >= ntrack || sect == 0 || sect > nsect || head >= nhead)
			return 0x140;

		int startsect = ((track * nhead) + head) * nsect + sect - 1;
		byte[] buf = new byte[n * 512];
		int ret = diskimage.read(startsect * 512, buf);
		if (ret == DiskImage.EIO)
			return 0x110;
		if (ret == DiskImage.EOF)
			return 0x104;
		mem.loadData(addr, buf);
 		return 0;
	}
	
    // Handle interrupt 0x13
    void int_disk(int[] reg, int[] sreg, int[] flags)
    {
        int ret;
        int ax = reg[Cpu.regAX];
        int bx = reg[Cpu.regBX];
        int cx = reg[Cpu.regCX];
        int dx = reg[Cpu.regDX];
        int es = sreg[Cpu.sregES];
        switch ((ax >> 8) & 0xff) {
          case 0x00:
            msg("int13: reset drive (dx=" + Misc.wordToHex(dx) + ")");
            ret = 0;
            break;
          case 0x02:
            ret = int13_readsectors(dx & 0xff, cx >> 8, cx & 0xff, dx >> 8, ax & 0xff, (es << 4) + bx);
            break;
          case 0x08:
            System.err.println("int13: get drive parameters (dx=" + Misc.wordToHex(dx) + ")");
            if ((dx & 0xff) != 0) {
                ret = 0x180;
                break;
            }
            long size = diskimage.getSize();
            dx = 0x0201;
            bx &= 0xff00;
            ret = 0;
            if (size == 360 * 1024) {
                bx |= 1;
                cx = 0x2809;
            } else if (size == 720 * 1024) {
                bx |= 3;
                cx = 0x5009;
            } else if (size == 1200 * 1024) {
                bx |= 2;
                cx = 0x500f;
            } else if (size == 1440 * 1024) {
                bx |= 4;
                cx = 0x5012;
            } else {
                System.err.println("Invalid disk size");
                ret = 0x107;
            }
            break;
          default:
            msg("int13: unknown function ax=" + Misc.wordToHex(ax));
            ret = 0x101;
            break;
        }
        flags[0] &= ~ Cpu.flCF;
        if ((ret & 0xff00) != 0)
            flags[0] |= Cpu.flCF;
        reg[Cpu.regAX] = (ax & 0x00ff) | ((ret << 8) & 0xff00);
    }


    public void handleInput(ExternalInputEvent evt)
    {
        if (evt instanceof KeyInputEvent)
            keyInputHandler.handleInput(evt);
    }

    public void runvm(String romfile, String diskfile)
    {
        // Create core simulation objects
        sched = new Scheduler();
        io = new IOPorts();
        mem = new Memory();
        cpu = new Cpu(sched, mem, io);
        sched.setCpu(cpu);

        // Create GUI window
        frame = new TerminalFrame("TestPC");
        frame.show();

        // Stop simulation when user closes the window
        frame.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                msg("Stopping simulation");
                sched.stopSimulation();
            }
        });

        // Interrupt controller
        intctl = new I8259(sched);
        io.registerHandler(intctl, 0x20, 2);
        cpu.setInterruptController(intctl);

        // Timer
        timer = new I8254(sched, intctl.getIrqLine(0));
        io.registerHandler(timer, 0x40, 4);

        // CGA card
        cga = new Cga(sched, mem, frame.getDisplay());
        io.registerHandler(cga, 0x3d0, 16);

        // Keyboard controller
        kbdctl = new KeyboardController(sched, intctl.getIrqLine(1));
        io.registerHandler(kbdctl, 0x60, 4);

        // Input handling
        keyInputHandler = new KeyInputHandler(kbdctl);
        sched.setInputHandler(this);
        frame.getDisplay().setInputHandler(sched);

        // Load ROM image into memory
        try {
            InputStream fin =
              TestPC.class.getClassLoader().getResourceAsStream(romfile);
            byte[] buf = new byte[65536];
            int i = fin.read(buf);
            fin.close();
            byte[] xbuf = new byte[i];
            System.arraycopy(buf, 0, xbuf, 0, i);
            mem.loadData(0x100000 - i, xbuf);
            System.err.println("Loaded " + i + " bytes from " + romfile);

            diskimage = new DiskImage(diskfile, true);

        } catch (IOException e) {
            e.printStackTrace();
            System.exit(1);
        }

        // Register I/O handlers
        iomisc = new IOMisc();
        io.registerHandler(iomisc, 0xe600, 4);

        // Register interrupt hooks
        cpu.setInterruptHook(this, 0xe6);

        // Go!
        try {
            sched.run();
        } catch (Cpu.InvalidOpcodeException e) {
            e.printStackTrace();
        }
    }

    public static void main(String[] args)
    {
        String romfile, diskfile;

        if (args.length == 1) {
            romfile = "bios/rombios.bin";
            diskfile = args[0];
        } else if (args.length == 2) {
            romfile = args[0];
            diskfile = args[1];
        } else {
            System.err.println("Usage: java TestPC [<romfile>] <diskfile>");
            System.exit(1);
            return;
        }

        TestPC a = new TestPC();
        a.runvm(romfile, diskfile);
    }

}

/* end */
